/*
    UHU-menu -- input plugin for freedesktop-style files
    Copyright (C) 2003, 2005  Koblinger Egmont <egmont@uhulinux.hu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _GNU_SOURCE
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>

#include "uhu-menu.h"

static struct item *item_new (void)
{
	struct item *item;
	item = malloc(sizeof(struct item));
	item->subitems_nr = 0;
	item->subitems = NULL;
	return item;
}

static int item_add_subitem (struct item *item, const struct item *subitem)
{
	if (item->subitems_nr == 0) {
		item->subitems = malloc(sizeof(struct item));
	} else {
		item->subitems = realloc(item->subitems, (item->subitems_nr + 1) * sizeof(struct item));
	}
	if (subitem != NULL) {
		memcpy(&item->subitems[item->subitems_nr], subitem, sizeof(struct item));
	} else {
		item->subitems[item->subitems_nr].subitems_nr = 0;
		item->subitems[item->subitems_nr].subitems = NULL;
	}
	item->subitems_nr++;
	return 0;
}

static int replace_item (struct item *dst, const struct item *src)
{
	struct item tmp;
	memcpy(&tmp, src, sizeof(struct item));
	tmp.issubmenu = dst->issubmenu;
	tmp.subitems = dst->subitems;
	tmp.subitems_nr = dst->subitems_nr;
	tmp.x_uhu_menu = dst->x_uhu_menu;
	memcpy(dst, &tmp, sizeof(struct item));
	return 0;
}

static int insert_directory_subitem (struct item *item, const struct item *subitem,
  const char *path)
{
	char buf[PATH_MAX];
	int i, j;

	buf[0] = '\0';
	path += strspn(path, "/");
	while (path[0] != '\0') {
		j = strcspn(path, "/");
		sprintf(buf + strlen(buf), "/%.*s", j, path);
		path += j;
		path += strspn(path, "/");
		for (i = 0; i < item->subitems_nr; i++) {
			if (!(strcmp(item->subitems[i].x_uhu_menu, buf))) {
				break;
			}
		}
		if (i == item->subitems_nr) {
			item_add_subitem(item, NULL);
			item->subitems[i].issubmenu = 1;
			item->subitems[i].x_uhu_menu = strdup(buf);
			item->subitems[i].sections = NULL;
		}
		item = &item->subitems[i];
	}
	if (item->sections != NULL) {
		message(3, "%s menu directory duplicated", buf[0] == '\0' ? "root" : buf);
		return -1;
	}
	replace_item(item, subitem);
	return 0;
}

static int insert_file_subitem (struct item *item, const struct item *subitem,
  const char *path)
{
	char buf[PATH_MAX];
	int i, j;

	buf[0] = '\0';
	path += strspn(path, "/");
	while (path[0] != '\0') {
		j = strcspn(path, "/");
		sprintf(buf + strlen(buf), "/%.*s", j, path);
		path += j;
		path += strspn(path, "/");
		for (i = 0; i < item->subitems_nr; i++) {
			if (!(strcmp(item->subitems[i].x_uhu_menu, buf))) {
				break;
			}
		}
		if (i == item->subitems_nr) {
			if (path[0] != '\0') {
				message(3, "unknown value for X-UHU-Menu");
				return -1;
			}
			item_add_subitem(item, subitem);
			item->subitems[i].issubmenu = 0;
			item->subitems[i].x_uhu_menu = strdup(buf);
			return 0;
		}
		item = &item->subitems[i];
	}
	return -1;
}

static char *get_x_uhu_menu_category (const struct item *item)
{
	char *cat;
	cat = get_value(item, NULL, "Categories", "");
	while (cat != NULL) {
		while (*cat == ';') cat++;
		if (!strncmp(cat, "X-UHU-Menu", 10) &&
		  (cat[10] == '/' || cat[10] == '\0' || cat[10] == ';')) {
			return cat + 10;
		}
		cat = strchr(cat, ';');
	}
	return NULL;
}

static void scan_directory_items (struct item *rootitem, const char *path, int path_prefix_len)
{
	struct item item;
	int fnlen;
	char *fn;
	char fn2[PATH_MAX];
	char buf[PATH_MAX];
	int pathlen;
	struct dirent *de;
	DIR *d;
	int l;

	d = opendir(path);
	if (d == NULL) {
		message_e(3, "opendir %s failed", path);
		return;
	}
	pathlen = sprintf(buf, "%s/", path);
	while ((de = readdir(d)) != NULL) {
		if (de->d_name[0] == '.') continue;

		strcpy(buf + pathlen, de->d_name);
		if (de->d_type == DT_UNKNOWN) {
			struct stat st;
			if (stat(buf, &st) < 0) {
				message_e(3, "stat %s failed", buf);
				continue;
			}
			if (S_ISDIR(st.st_mode)) {
				de->d_type = DT_DIR;
			}
		}
		if (de->d_type == DT_DIR) {
			scan_directory_items(rootitem, buf, path_prefix_len);
			continue;
		}

		l = strlen(de->d_name);
		if (l < 10 || strcmp(de->d_name + l - 10, ".directory")) continue;

		strcpy(buf + pathlen, de->d_name);
		if (desktop_parse(buf, path_prefix_len, &item) >= 0) {
			fn = get_value(&item, NULL, "Type", "");
			if (fn != NULL && !strcmp(fn, "Directory")) {
				fn = get_x_uhu_menu_category(&item);
				if (fn != NULL) {
					fnlen = strcspn(fn, ";");
					sprintf(fn2, "%.*s", fnlen, fn);
					insert_directory_subitem(rootitem, &item, fn2);
				}
			}
		}
	}
	closedir(d);
}

static void scan_file_items (struct item *rootitem, const char *path, int path_prefix_len)
{
	struct item item;
	int fnlen;
	char *fn;
	char fn2[PATH_MAX];
	char buf[PATH_MAX];
	int pathlen;
	struct dirent *de;
	DIR *d;
	int l;

	d = opendir(path);
	if (d == NULL) {
		message_e(3, "opendir %s failed", path);
		return;
	}
	pathlen = sprintf(buf, "%s/", path);
	while ((de = readdir(d)) != NULL) {
		if (de->d_name[0] == '.') continue;

		strcpy(buf + pathlen, de->d_name);
		if (de->d_type == DT_UNKNOWN) {
			struct stat st;
			if (stat(buf, &st) < 0) {
				message_e(3, "stat %s failed", buf);
				continue;
			}
			if (S_ISDIR(st.st_mode)) {
				de->d_type = DT_DIR;
			}
		}
		if (de->d_type == DT_DIR) {
			scan_file_items(rootitem, buf, path_prefix_len);
			continue;
		}

		l = strlen(de->d_name);
		if (l < 8 || strcmp(de->d_name + l - 8, ".desktop")) continue;

		strcpy(buf + pathlen, de->d_name);
		if (desktop_parse(buf, path_prefix_len, &item) >= 0) {
			fn = get_value(&item, NULL, "Type", "");
			if (fn != NULL && !strcmp(fn, "Application")) {
				fn = get_x_uhu_menu_category(&item);
				if (fn != NULL) {
					fnlen = strcspn(fn, ";");
					sprintf(fn2, "%.*s/%s", fnlen, fn, de->d_name);
					insert_file_subitem(rootitem, &item, fn2);
				}
			}
		}
	}
	closedir(d);
}

struct item *read_menu (const char *path)
{
	struct item *rootitem;
	char buf[PATH_MAX];
	int path_prefix_len;

	rootitem = item_new();
	rootitem->issubmenu = 1;
	rootitem->x_uhu_menu = "";
	rootitem->sections = NULL;

	sprintf(buf, "%s/desktop-directories", path);
	path_prefix_len = strlen(buf) + 1;
	scan_directory_items(rootitem, buf, path_prefix_len);

	sprintf(buf, "%s/applications", path);
	path_prefix_len = strlen(buf) + 1;
	scan_file_items(rootitem, buf, path_prefix_len);

	return rootitem;
}
